import os
from typing import Any, List, Optional, Union

from ..globals import DEFAULT_MYKEFILE
from ..io.read import read
from ..utils import make_executable


class write:
    def __new__(  # type: ignore
        cls,
        content: Union[str, bytes],
        path: str,
        append: bool = False,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        cls.text(
            content=content,
            path=path,
            append=append,
            overwrite=overwrite,
            **kwargs,
        )

    @staticmethod
    def text(
        content: Union[str, bytes],
        path: str,
        append: bool = False,
        overwrite: bool = False,
        encoding: str = "utf-8",
        **kwargs: Any,
    ) -> None:
        mode_default: str = "w"

        mode: str = kwargs.pop("mode", mode_default)

        if os.path.exists(path):
            if overwrite:
                os.remove(path)
            elif append and mode == mode_default:
                mode = "a"
            else:
                raise FileExistsError(path)

        if mode == mode_default:
            if isinstance(content, bytes):
                mode += "b"

            if not os.path.exists(path):
                mode += "+"

        with open(path, mode=mode, encoding=encoding, **kwargs) as f:
            f.write(content)

    @classmethod
    def lines(
        cls,
        content: List[Optional[str]],
        path: str,
        append: bool = False,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        if isinstance(content, str):
            raise TypeError("expected a list of strings, not a string")
        cls.text(
            path=path,
            content=os.linesep.join([str(x) for x in content]),
            append=append,
            overwrite=overwrite,
            **kwargs,
        )

    @classmethod
    def mykefile(cls, path: Optional[str] = None) -> None:
        if not path:
            path = os.getcwd()

        if os.path.isdir(path):
            path = os.path.join(path, DEFAULT_MYKEFILE)

        cls.text(
            path=path,
            content="""#!/usr/bin/env python3
# type: ignore
# pylint: skip-file
# flake8: noqa
import myke  # noqa

myke.require(
    pip_args=[
        "--extra-index-url",
        "https://codeberg.org/api/packages/Fresh2dev/pypi/simple",
    ],
    mykefiles="0.0.1a3.dev44",
)
from mykefiles import hello_world  # noqa

# @myke.task(root=True)
# def setup():
#     # setup
#     ...
#
#     yield
#
#     # teardown
#     ...


if __name__ == "__main__":
    myke.main(__file__)
""",
        )

        make_executable(path)

    @classmethod
    def url(
        cls,
        addr: str,
        path: Optional[str] = None,
        append: bool = False,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> str:
        if not path:
            path = os.getcwd()

        if os.path.isdir(path):
            path = os.path.join(path, os.path.basename(addr))

        resp: str = read.url(addr)

        cls.text(content=resp, path=path, append=append, overwrite=overwrite, **kwargs)

        return path



import os

import pytest


def test_import_module(clean_dir):
    import myke

    myke.TASKS.clear()

    file_name: str = "py_pkg.py"
    expected_file: str = os.path.join(os.getcwd(), "tasks", file_name)

    myke.import_module(
        "https://codeberg.org/fresh2dev/mykefiles/raw/branch/dev/src/mykefiles/"
        + file_name,
    )
    with pytest.raises(FileExistsError):
        myke.install_module(
            "https://codeberg.org/fresh2dev/mykefiles/raw/branch/dev/src/mykefiles/"
            + file_name,
            fail_if_exists=True,
        )

    os.environ["MYKE_UPDATE_MODULES"] = "1"

    myke.install_module(
        "https://codeberg.org/fresh2dev/mykefiles/raw/branch/dev/src/mykefiles/"
        + file_name,
        fail_if_exists=True,
    )

    assert os.path.exists(expected_file)


def test_install_module_notfound(clean_dir):
    import myke

    myke.TASKS.clear()

    expected_file: str = os.path.join(os.getcwd(), "tasks", "LICENSE")

    with pytest.raises(myke.exceptions.MykeNotFoundError):
        myke.install_module(
            "https://raw.githubusercontent.com/fresh2dev/AnyBox/main/LICENSE",
        )

    assert not os.path.exists(expected_file)



from __future__ import annotations

import subprocess


class MykeNotFoundError(Exception):
    ...


class TaskAlreadyRegisteredError(Exception):
    ...


class NoTasksFoundError(Exception):
    ...


class CalledProcessError(subprocess.CalledProcessError):
    def __str__(self) -> str:
        if self.returncode and self.returncode < 0:
            return super().__str__()
        else:
            msg: str = (
                f"Command '{self.cmd}' returned non-zero exit status {self.returncode}:"
            )
            for attr in "stdout", "stderr":
                attr_value: str = getattr(self, attr)
                if attr_value:
                    attr_value_trimmed: str = attr_value.rstrip("\n")
                    if attr_value_trimmed:
                        msg += f"\n{attr}:\n{attr_value_trimmed}"
            return msg



import os
import subprocess
from typing import List, Union
from unittest import mock
from uuid import uuid4

import pytest

from myke import sh_stdout


def _common(
    capsys,
    resources_dir: str,
    cli_args: List[str],
    expected_txt: Union[str, List[str]],
):
    # 1. ARRANGE
    import myke
    from myke.main import sys as target_sys

    myke.TASKS.clear()

    mykefile: str = os.path.join(resources_dir, "Mykefile")

    # 2. ACT
    myke.import_module(mykefile)

    with mock.patch.object(target_sys, "argv", [""] + cli_args):
        myke.main(mykefile)

    # 3. ASSERT
    captured = capsys.readouterr()

    if expected_txt:
        if isinstance(expected_txt, str):
            assert expected_txt in captured.out
        else:
            for exp in expected_txt:
                assert exp in captured.out

    return captured


def test_Mykefile_setup(capsys, resources_dir: str):
    expected_txt = "charging lazer beamz"
    cli_args: List[str] = ["hello", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_named(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello-named", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_named2(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello-named2", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_named3(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello-named3", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_positional(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello-positional", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_positional_default(capsys, resources_dir: str):
    expected_txt = "hello world"
    cli_args: List[str] = ["hello-positional"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_positional_bool_true(capsys, resources_dir: str):
    expected_txt = "hello world".upper()
    cli_args: List[str] = ["hello-positional-bool", "y"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_positional_bool_false(capsys, resources_dir: str):
    expected_txt = "hello world"
    cli_args: List[str] = ["hello-positional-bool", "n"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_sh(capsys, resources_dir: str):
    txt = str(uuid4())
    cli_args: List[str] = ["hello-sh-stdout", "--name", txt]
    expected_txt: List[str] = [txt, "woohoo"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_env(capsys, resources_dir: str):
    expected_txt = "donald"
    cli_args: List[str] = ["hello-env"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_env2(clean_dir, capsys, resources_dir: str):
    from myke import write

    expected_txt: str = str(uuid4())
    write(f"NAME={expected_txt}", path=".env")
    cli_args: List[str] = ["--myke-env-file", ".env", "hello-env2"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_env_update(capsys, resources_dir: str):
    expected_txt = "donald"
    cli_args: List[str] = ["hello-env-update"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_depends(capsys, resources_dir: str):
    txt = str(uuid4())
    cli_args: List[str] = ["hello-depends", "--name", txt]
    expected_txt: List[str] = [txt, "woohoo"]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_othermyke(capsys, resources_dir: str):
    expected_txt = str(uuid4())
    cli_args: List[str] = ["hello-othermyke", "--name", expected_txt]
    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=expected_txt)


def test_Mykefile_hello_check(capsys, resources_dir: str):
    cli_args: List[str] = ["hello-check"]

    with pytest.raises(subprocess.CalledProcessError):
        _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt="...")


def test_Mykefile_hello_timeout(capsys, resources_dir: str):
    cli_args: List[str] = ["hello-timeout"]

    with pytest.raises(subprocess.TimeoutExpired):
        _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt="...")


def test_Mykefile_assert_types(capsys, resources_dir: str):
    cli_args: List[str] = [
        "assert-types",
        "--xstr",
        "hello",
        "--xint",
        "5",
        "--xfloat",
        "3.14",
        "--xbool",
        "--xliststr",
        "list[this that]",
        "list[another]",
        "--xsetint",
        "list[1 1]",
        "list[1 2]",
        "list[  3  ]",
        "--xtuplefloat",
        "list[3.14]",
        "list[6.28 9.42]",
    ]

    _ = _common(capsys, resources_dir, cli_args=cli_args, expected_txt=None)


def test_Mykefile_executable(capsys, resources_dir: str):
    # 1. ARRANGE
    mykefile: str = os.path.join(resources_dir, "Mykefile")

    expected_txt: str = str(uuid4())

    # 2. ACT
    output: str = sh_stdout(
        f"""
    PYTHONPATH=./src python -m myke --myke-file {mykefile} hello --name {expected_txt}
""",
        echo=True,
    )

    # 3. ASSERT
    assert output
    assert expected_txt in output

    captured = capsys.readouterr()
    assert captured.out
    assert not captured.err
    assert expected_txt in captured.out



import os
import sys
from dataclasses import dataclass
from fnmatch import fnmatch
from inspect import getsource
from typing import Any, Callable, List, Optional

import yapx

from .__version__ import __version__
from .exceptions import NoTasksFoundError
from .globals import DEFAULT_MYKEFILE, MYKE_VAR_NAME, ROOT_TASK_KEY, TASKS
from .io.echo import echo
from .io.read import read
from .io.write import write
from .tasks import import_module

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

__all__ = ["__version__", "main", "sys"]


def main(_file: Optional[str] = None) -> None:
    @dataclass
    class MykeArgs(yapx.types.Dataclass):
        file: Annotated[
            str,
            yapx.arg(
                default=_file if _file else DEFAULT_MYKEFILE,
                flags=["--myke-file"],
                env="MYKE_FILE",
                group="myke args",
            ),
        ]
        file_paths: Annotated[
            List[str],
            yapx.arg(
                default=lambda: [os.path.expanduser("~"), os.getcwd()],
                flags=["--myke-file-paths"],
                env="MYKE_FILE_PATHS",
                group="myke args",
            ),
        ]
        env_file: Annotated[
            Optional[str],
            yapx.arg(
                default=None,
                flags=["--myke-env-file"],
                env="MYKE_ENV_FILE",
                group="myke args",
            ),
        ]
        update_modules: Annotated[
            bool,
            yapx.arg(
                default=False,
                flags=["--myke-update-modules"],
                env="MYKE_UPDATE_MODULES",
                group="myke args",
            ),
        ]
        task_help: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["-h", "--help"],
            ),
        ]
        task_help_full: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["--help-full"],
            ),
        ]
        help: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["--myke-help"],
            ),
        ]
        explain: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["--myke-explain"],
            ),
        ]
        version: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["--myke-version"],
            ),
        ]
        create: Annotated[
            Optional[bool],
            yapx.arg(
                default=None,
                group="myke args",
                exclusive=True,
                flags=["--myke-create"],
            ),
        ]

    prog: str = _file if _file else MYKE_VAR_NAME

    parser = yapx.ArgumentParser(
        prog=prog,
        add_help=False,
        description="Parameters for Myke:",
    )
    parser.add_arguments(MykeArgs)

    myke_args: MykeArgs
    task_args: List[str]
    myke_args, task_args = parser.parse_known_args_to_model(
        sys.argv[1:],
        args_model=MykeArgs,
        skip_pydantic_validation=True,
    )
    assert isinstance(myke_args, MykeArgs)

    if myke_args.version:
        echo(__version__)
        parser.exit()
    elif myke_args.create:
        write.mykefile(myke_args.file)
        echo(f"Created: {myke_args.file}")
        parser.exit()

    if myke_args.env_file:
        os.environ.update(read.envfile(myke_args.env_file))

    if myke_args.update_modules:
        os.environ["MYKE_UPDATE_MODULES"] = "1"

    if _file:
        if not os.path.exists(_file):
            raise FileNotFoundError(_file)
        if not TASKS:
            raise NoTasksFoundError(_file)
        _file = os.path.abspath(_file)

    mykefiles: List[str] = (
        [myke_args.file]
        if os.path.dirname(myke_args.file)
        else [
            y
            for x in myke_args.file_paths
            for y in [os.path.join(x, myke_args.file)]
            if os.path.exists(y)
        ]
    )

    for f in mykefiles:
        f = os.path.abspath(f)
        try:
            if f != _file:
                import_module(f)
            elif not TASKS:
                raise NoTasksFoundError(f)
        except FileNotFoundError:
            parser.print_help()
            echo(
                (
                    f"{os.linesep}"
                    f"'{myke_args.file}' not found. Create it using:"
                    f"{os.linesep}"
                    f"> {prog} --myke-create --myke-file '{f}'"
                    f"{os.linesep}"
                ),
            )
            parser.exit()

    root_task: Optional[Callable[..., Any]] = TASKS.pop(ROOT_TASK_KEY, None)

    if myke_args.explain:
        explain_this: Optional[Callable[..., Any]] = None

        if not task_args or task_args[0].startswith("-"):
            explain_this = root_task
            if explain_this is None:
                echo("There is no root task. Provide a task name to explain.")
        else:
            explain_this = TASKS.get(task_args[0], None)
            if explain_this is None:
                echo(f"Given task name not found: {task_args[0]}")

        if explain_this:
            echo(getsource(explain_this))

        parser.exit()

    if myke_args.help or (not task_args and not myke_args.task_help_full):
        parser.print_help()
        echo.tasks(prog=prog)
        parser.exit()

    if task_args and "*" in task_args[0]:
        for k in list(TASKS):
            if not fnmatch(k, task_args[0]):
                del TASKS[k]
        if not myke_args.task_help_full:
            echo.tasks(prog=prog)
            parser.exit()

    if myke_args.task_help:
        task_args.append("--help")

    yapx.run(
        root_task,
        _args=task_args,
        _prog=prog,
        _print_help=myke_args.task_help_full,
        **TASKS,
    )



import collections.abc
import os
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Union

from ..globals import MYKE_VAR_NAME, TASKS


class echo:
    def __new__(cls, *args: Any, **kwargs: Any) -> None:  # type: ignore
        cls.text(*args, **kwargs)

    @staticmethod
    def _print(*args: Any, print_kwargs: Optional[Dict[str, Any]] = None) -> None:
        if not print_kwargs:
            print_kwargs = {}
        print(*args, **print_kwargs)

    @classmethod
    def text(cls, *args: Any, **kwargs: Any) -> None:
        cls._print(*args, print_kwargs=kwargs)

    @classmethod
    def lines(
        cls,
        seq: Sequence[Optional[str]],
        linesep: str = os.linesep,
        print_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        cls._print(linesep.join([str(x) for x in seq]), print_kwargs=print_kwargs)

    @classmethod
    def json(
        cls,
        obj: Any,
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        import json as _json

        cls._print(_json.dumps(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def table(
        cls,
        obj: Union[Iterable[Iterable[Any]], Mapping[str, Iterable[Any]]],
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        from tabulate import tabulate

        if (
            obj
            and isinstance(obj, collections.abc.Sequence)
            and isinstance(obj[0], dict)
            and "headers" not in kwargs
        ):
            kwargs["headers"] = "keys"

        cls._print(tabulate(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def pretty(
        cls,
        obj: Any,
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        from pprint import pformat

        cls._print(pformat(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def tasks(cls, prog: Optional[str] = None, tablefmt: Optional[str] = None) -> None:
        cls.text()

        if not TASKS:
            cls.text("No tasks found.")
        else:
            records: List[Dict[str, str]] = [
                {"Task": k, "Source": v.__module__} for k, v in sorted(TASKS.items())
            ]

            if not prog:
                prog = MYKE_VAR_NAME

            try:
                if not tablefmt:
                    tablefmt = "rst"
                echo.table(records, tablefmt=tablefmt)
            except ModuleNotFoundError:
                cls.text("TASKS")
                cls.text("-----")
                cls.lines([str(x) if x else "" for x in records])

            cls.text()
            cls.text("To view task parameters, see:")
            cls.text(f"> {prog} <task-name> --help")

        cls.text()



from __future__ import annotations

import sys
from typing import Callable

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class MykeType(Protocol):
    TASKS: dict[str, Callable[..., None]]



import os
from pathlib import Path
from typing import Any, Dict, List

import pytest

import myke


def test_read_text(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.txt")

    # 2. ACT
    content: str = myke.read(file)

    # 3. ASSERT
    assert content
    assert isinstance(content, str)
    assert content.startswith("hello")
    assert content.endswith("world")


def test_read_lines(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.txt")

    # 2. ACT
    content: List[str] = myke.read.lines(file)

    # 3. ASSERT
    assert content
    assert content == ["hello", "world"]


def test_read_json(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.json")

    # 2. ACT
    content: Dict[str, Any] = myke.read.json(file)

    # 3. ASSERT
    assert content
    assert content == {"hello": "world"}


def test_read_yaml(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.yaml")

    # 2. ACT
    content: Dict[str, Any] = myke.read.yaml(file)

    # 3. ASSERT
    assert content
    assert content == {"hello": "world"}


def test_read_yaml_all(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test-all.yaml")

    # 2. ACT
    content: Dict[str, Any] = myke.read.yaml_all(file)

    # 3. ASSERT
    assert content
    assert content == [{"hello": "world"}, {"goodbye": "cruel world"}]


def test_read_toml(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.toml")

    # 2. ACT
    content: Dict[str, Any] = myke.read.toml(file)

    # 3. ASSERT
    assert content
    assert content == {
        "arrival": {"hello": "world"},
        "departure": {"goodbye": "cruel world"},
    }


def test_read_cfg(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.cfg")

    # 2. ACT
    content: Dict[str, Any] = myke.read.cfg(file)

    # 3. ASSERT
    assert content
    assert content == {
        "arrival": {"hello": "world"},
        "departure": {"goodbye": "cruel world"},
    }
    assert myke.read.ini(file) == content


def test_read_dotfile(resources_dir: str):
    # 1. ARRANGE
    file: str = os.path.join(resources_dir, "files", "test.env")

    # 2. ACT
    content: Dict[str, Any] = myke.read.dotfile(file)

    # 3. ASSERT
    assert content
    assert content == {"hello": "world", "goodbye": "cruel world"}
    assert myke.read.envfile(file) == content


def test_read_url():
    url: str = (
        "https://codeberg.org/fresh2dev/copier-f2dv-project/raw/branch/main/LICENSE"
    )

    resp: str = myke.read.url(url)

    assert resp
    assert isinstance(resp, str)


def test_write_text(tmp_path: Path):
    expected: str = "hello world"

    path = str(tmp_path / "dummy.txt")
    myke.write(expected, path)

    assert os.path.exists(path)
    assert myke.read(path) == expected

    myke.write(expected, path, append=True)
    assert myke.read(path) == expected * 2

    with pytest.raises(FileExistsError):
        myke.write(content=expected, path=path)


def test_write_lines(tmp_path: Path):
    expected: List[str] = ["hello", "world"]

    path = tmp_path / "dummy.txt"
    myke.write.lines(path=path, content=expected)

    assert path.exists()
    assert myke.read.lines(path) == expected

    with pytest.raises(TypeError):
        myke.write.lines(path=path, content=str(expected))


def test_echo_text(capsys):
    test_input: str = "hello world"
    expected: str = test_input + os.linesep

    myke.echo(test_input)

    captured = capsys.readouterr()
    assert captured.out == expected


def test_echo_lines(capsys):
    test_input: List[str] = ["hello", None, "world", None]
    expected: str = os.linesep.join([str(x) for x in test_input]) + os.linesep

    myke.echo.lines(test_input)

    captured = capsys.readouterr()
    assert captured.out == expected


def test_echo_json(capsys):
    test_input: List[str] = ["hello", "world"]
    expected: str = '["hello", "world"]' + os.linesep

    myke.echo.json(test_input)

    captured = capsys.readouterr()
    assert captured.out == expected


def test_echo_table(capsys):
    test_input: List[str] = [{"hello": "world"}, {"hello": "world2"}]
    expected: str = "hello\n-------\nworld\nworld2\n"

    myke.echo.table(test_input)

    captured = capsys.readouterr()
    assert captured.out == expected


def test_echo_pretty(capsys):
    test_input: str = "hello world"
    expected: str = f"'{test_input}'" + os.linesep

    myke.echo.pretty(test_input)

    captured = capsys.readouterr()
    assert captured.out == expected


def test_echo_tasks(capsys):
    myke.TASKS.clear()
    expected: str = "No tasks found."

    myke.echo.tasks()

    captured = capsys.readouterr()
    assert expected in captured.out



import os
from typing import List, Pattern
from unittest import mock

import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import myke
from myke.main import sys as target_sys


def test_main_version(capsys: CaptureFixture, version_regex: Pattern):
    # 1. ARRANGE
    args: List[str] = ["--myke-version"]

    # 2. ACT
    with mock.patch.object(target_sys, "argv", [""] + args), pytest.raises(
        SystemExit,
    ) as pexit:
        myke.main()

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert version_regex.search(captured.out), "invalid version"
    assert pexit.value.code == 0


def test_main_help(capsys: CaptureFixture):
    # 1. ARRANGE
    args: List[str] = ["--myke-help"]

    # 2. ACT
    with mock.patch.object(target_sys, "argv", [""] + args), pytest.raises(
        SystemExit,
    ) as pexit:
        myke.main()

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert pexit.value.code == 0


def test_main_explain(capsys: CaptureFixture, resources_dir: str):
    # 1. ARRANGE
    args: List[str] = ["--myke-explain", "hello"]

    myke.TASKS.clear()
    mykefile: str = os.path.join(resources_dir, "Mykefile")

    expected: str = '''
@myke.task\ndef hello(name="world"):\n    """Say hello."""\n    print("hello " + name)
'''

    # 2. ACT
    myke.import_module(mykefile)

    with mock.patch.object(target_sys, "argv", [""] + args), pytest.raises(
        SystemExit,
    ) as pexit:
        myke.main(mykefile)

    # 3. ASSERT
    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert pexit.value.code == 0

    assert expected.strip() == captured.out.strip()



import subprocess
from typing import List

import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import myke


def test_run(capfd: CaptureFixture):
    expected: str = "hello world"
    stdout, stderr, returncode = myke.run(["python", "-c", f"print('{expected}')"])

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert captured.out.rstrip() == expected

    assert returncode == 0
    assert not stderr
    assert not stdout


def test_run_shell(capfd: CaptureFixture):
    expected: str = "hello world"
    stdout, stderr, returncode = myke.run(f'echo "{expected}"')

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert captured.out.rstrip() == expected

    assert returncode == 0
    assert not stderr
    assert not stdout


def test_run_no_echo(capfd: CaptureFixture):
    expected: str = "hello world"
    stdout, stderr, returncode = myke.run(f'echo "{expected}"', echo=False)

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert returncode == 0
    assert not stderr
    assert not stdout


def test_run_capture_no_echo(capfd: CaptureFixture):
    expected: str = "hello world"
    stdout, stderr, returncode = myke.run(
        f'echo "{expected}"',
        capture_output=True,
        echo=False,
    )

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert returncode == 0
    assert not stderr
    assert stdout.rstrip() == expected


def test_run_check():
    with pytest.raises(subprocess.CalledProcessError):
        myke.run("exit 1")


def test_run_no_check():
    expected: int = 1
    _, _, returncode = myke.run(f"exit {expected}", check=False)
    assert returncode == expected


def test_sh():
    cmd: str = "echo hello"
    assert myke.run(cmd, shell=True) == myke.sh(cmd)


def test_sh_stdout_lines():
    expected: List[str] = ["hello", "world"]
    cmd: str = "echo '" + "\\n".join(expected) + "'"
    stdout = myke.sh_stdout_lines(cmd)

    assert stdout == expected


def test_sh_stdout():
    expected: List[str] = ["hello", "world"]
    cmd: str = "echo '" + "\\n".join(expected) + "'"
    stdout = myke.sh_stdout(cmd)

    assert stdout == "\n".join(expected)


def test_require(capfd: CaptureFixture):
    _stdout, stderr, returncode = myke.require(
        pip_args=[
            "--extra-index-url",
            "https://codeberg.org/api/packages/Fresh2dev/pypi/simple",
            "--quiet",
            "--dry-run",
        ],
        mykefiles="0.0.1a3.dev44",
    )

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert returncode == 0
    assert not stderr



from __future__ import annotations

import json
import os
import subprocess
import sys
from functools import wraps
from typing import Any, Sequence

from .exceptions import CalledProcessError
from .utils import split_and_trim_text

__all__ = ["run", "sh", "sh_stdout", "sh_stdout_lines", "require"]


def run(
    args: str | Sequence[str],
    capture_output: None | bool = False,
    echo: bool | None = True,
    check: bool | None = True,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    env_update: dict[str, str | None] | None = None,
    timeout: float | None = None,
    shell: bool | None = None,
    **kwargs: Any,
) -> tuple[str | None, str | None, int]:
    if shell is None:
        shell = isinstance(args, str) and " " in args

    env = env.copy() if env else os.environ.copy()

    if env_update:
        for k, v in env_update.items():
            if v is None:
                env.pop(k, None)
            else:
                env[k] = v

    if not echo and not capture_output:
        for k in ("stdout", "stderr"):
            kwargs[k] = subprocess.DEVNULL

    p: subprocess.CompletedProcess[str] = subprocess.run(
        args,
        shell=shell,
        env=env,
        cwd=cwd,
        timeout=timeout,
        capture_output=bool(capture_output),
        text=capture_output,
        check=False,
        **kwargs,
    )

    if echo and capture_output:
        for out in (p.stdout, p.stderr):
            if out:
                print(out.rstrip(os.linesep))

    if check and p.returncode:
        raise CalledProcessError(p.returncode, p.args, p.stdout, p.stderr)

    return p.stdout, p.stderr, p.returncode


@wraps(run)
def sh(*args: Any, **kwargs: Any) -> tuple[str | None, str | None, int]:
    kwargs["shell"] = True
    return run(*args, **kwargs)


@wraps(sh)
def sh_stdout_lines(*args: Any, **kwargs: Any) -> list[str]:
    kwargs["capture_output"] = True
    kwargs["echo"] = kwargs.get("echo", False)
    stdout, *_ = sh(*args, **kwargs)
    return split_and_trim_text(stdout)


@wraps(sh)
def sh_stdout(*args: Any, **kwargs: Any) -> str:
    return os.linesep.join(sh_stdout_lines(*args, **kwargs))


def _run_pip(
    *args: str,
    pip_args: list[str] | None = None,
    echo: bool = False,
    capture_output: bool = False,
    **kwargs: str,
) -> tuple[str | None, str | None, int]:
    if not pip_args:
        pip_args = []

    return run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            *pip_args,
            *args,
            *[f"{k}=={v}" for k, v in kwargs.items()],
        ],
        echo=echo,
        capture_output=capture_output,
    )


def require(
    *args: str,
    pip_args: list[str] | None = None,
    skip_check: bool = False,
    **kwargs: str,
) -> tuple[str | None, str | None, int]:
    if not pip_args:
        pip_args = []

    dry_run_args: list[str] = (
        [] if skip_check else ["-qq", "--dry-run", "--report", "-"]
    )

    stdout, stderr, returncode = _run_pip(
        *args,
        pip_args=pip_args + dry_run_args,
        echo=False,
        capture_output=True,
        **kwargs,
    )

    if dry_run_args:
        if not stdout or returncode != 0:
            print(stdout, stderr, returncode)
            return stdout, stderr, returncode

        result: dict[str, Any] = json.loads(stdout)

        if result and result.get("install"):
            return _run_pip(
                *args,
                pip_args=pip_args,
                echo=True,
                capture_output=False,
                **kwargs,
            )

    return stdout, stderr, returncode



from __future__ import annotations

from typing import Callable

TASKS: dict[str, Callable[..., None]] = {}

MYKE_VAR_NAME: str = "myke"

DEFAULT_MYKEFILE: str = "Mykefile"

ROOT_TASK_KEY: str = "__root__"



import sys

# import urllib.request
from functools import partial, wraps
from typing import Any, Callable, Dict, List

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard


class read(str):
    def __new__(cls, path: str, encoding: str = "utf-8") -> str:  # type: ignore
        return cls.text(path=path, encoding=encoding)

    @staticmethod
    def _is_simple_dict(candidate: Any) -> TypeGuard[Dict[str, Any]]:
        return isinstance(candidate, dict) and all(
            isinstance(k, str) for k in candidate
        )

    @classmethod
    def _read_simple_dict(
        cls,
        reader: Callable[..., Any],
        **kwargs: Any,
    ) -> Dict[str, Any]:
        content: Any = reader(**kwargs)
        if not cls._is_simple_dict(content):
            raise TypeError("expected a dictionary with string keys")
        return content

    @staticmethod
    def text(path: str, encoding: str = "utf-8") -> str:
        with open(path, encoding=encoding) as f:
            return f.read().strip()

    @classmethod
    @wraps(text)
    def lines(cls, *args: str, **kwargs: str) -> List[str]:
        return [
            y for x in cls.text(*args, **kwargs).splitlines() for y in [x.strip()] if y
        ]

    @classmethod
    @wraps(text)
    def json(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        import json as _json

        return cls._read_simple_dict(partial(_json.loads, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(text)
    def yaml(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        import yaml as _yaml

        return cls._read_simple_dict(
            partial(_yaml.safe_load, cls.text(*args, **kwargs)),
        )

    @classmethod
    @wraps(text)
    def yaml_all(cls, *args: str, **kwargs: str) -> List[Dict[str, Any]]:
        import yaml as _yaml

        def _yaml_all(txt: str) -> List[Dict[str, Any]]:
            return [
                cls._read_simple_dict(lambda y: y, y=x)
                for x in _yaml.safe_load_all(txt)
            ]

        return _yaml_all(cls.text(*args, **kwargs))

    @classmethod
    @wraps(text)
    def toml(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        try:
            import tomli as _toml
        except ImportError:
            import tomllib as _toml

        return cls._read_simple_dict(partial(_toml.loads, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(text)
    def cfg(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        from configparser import ConfigParser

        def _read_cfg(txt: str) -> Dict[str, Any]:
            cp = ConfigParser()
            cp.read_string(txt)
            return {x: dict(cp.items(x)) for x in cp.sections()}

        return cls._read_simple_dict(partial(_read_cfg, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(cfg)
    def ini(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        return cls.cfg(*args, **kwargs)

    @classmethod
    @wraps(text)
    def dotfile(cls, *args: str, **kwargs: str) -> Dict[str, str]:
        from io import StringIO

        from dotenv import dotenv_values

        return cls._read_simple_dict(
            partial(dotenv_values, stream=StringIO(cls.text(*args, **kwargs))),
        )

    @classmethod
    @wraps(dotfile)
    def envfile(cls, *args: str, **kwargs: str) -> Dict[str, str]:
        return cls.dotfile(*args, **kwargs)

    @staticmethod
    def _url(addr: str, **kwargs: Any) -> Any:
        import requests

        addr = kwargs.pop("url", addr)
        method: str = kwargs.pop("method", "GET")
        timeout: float = kwargs.pop("timeout", 10)
        resp: requests.Response = requests.request(
            method=method,
            url=addr,
            timeout=timeout,
            **kwargs,
        )
        return resp

    @classmethod
    def url(cls, addr: str, **kwargs: Any) -> str:
        resp_text: str = cls._url(addr=addr, **kwargs).text
        return resp_text

    @classmethod
    def url_dict(cls, addr: str, **kwargs: Any) -> Dict[str, Any]:
        resp: Any = cls._url(addr=addr, **kwargs).json()
        resp_dict: Dict[str, Any] = cls._read_simple_dict(lambda: resp)
        return resp_dict
